/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#ifndef __SparseArray2D_hh__
#define __SparseArray2D_hh__

#include <bias_config.h>

#include <Debug/Exception.hh>
#include <Debug/Error.hh>

#include <map>

namespace BIAS {
  
  /** @class SparseArray2D
      @brief generic two dimensional psarsly populated rectangular array 
      holding arbitrary data types

      The syntax of the access functions is kept similar to the stl:: classes
      
      @author woelk 11/2007 (c) www.vision-n.de */
  template <class T>  
  class  SparseArray2D 
  {
  public:
    SparseArray2D();

    SparseArray2D(const unsigned nrows, const unsigned ncols);

    // copy constructor
    SparseArray2D(const SparseArray2D<T>& m);

    ~SparseArray2D();

    /// copy operator
    SparseArray2D<T>& operator=(const SparseArray2D<T>& m);

    /// frees the memory
    void clear();

    /// preserves the content
    void resize(const unsigned  nrows, const unsigned ncols);

    bool empty() const 
    { return Data_.empty(); }

    unsigned ncols() const { return Ncols_; }

    unsigned nrows() const { return Nrows_; }

    unsigned size() const { return Nrows_ * Ncols_; }

    unsigned num_entries() const { return Data_.size(); }

    bool is_valid(const unsigned row, const unsigned col) const;

    void erase(const unsigned row, const unsigned col);

    /// checked element access
    T&       operator()(const unsigned row, const unsigned col);
    const T& operator()(const unsigned row, const unsigned col) const;

    /// unchecked element access
    // T&       operator[](const unsigned row, const unsigned col) 
//     { return Data_[Indices2Map(row, col)]; }
//     const T& operator[](const unsigned row, const unsigned col) const 
//     { return Data_[Indices2Map(row, col)]; }

    typedef typename std::map<long long unsigned, T>::const_iterator const_iterator;
    typedef typename std::map<long long unsigned, T>::iterator iterator;

    const_iterator begin() const { return Data_.begin(); }
    const_iterator end() const { return Data_.end(); }
    iterator begin() { return Data_.begin(); }
    iterator end() { return Data_.end(); }
    
    inline long long unsigned Indices2Map(const unsigned row, 
                                          const unsigned col) const
    { return (long long unsigned)row * (long long unsigned)Ncols_ + 
        (long long unsigned)col; }

    inline void Map2Indices(const long long unsigned& map,
                            unsigned& row, unsigned & col) const
    { row = (unsigned)(map / (long long unsigned)Ncols_); 
      col = (unsigned)(map % (long long unsigned)Ncols_); }

  private:
    std::map<long long unsigned, T> Data_;
    unsigned Nrows_, Ncols_;


  };

  /////////////////////////////////////////////////////////////////
  // implementation
  /////////////////////////////////////////////////////////////////

  template <class T>
  SparseArray2D<T>::SparseArray2D()
    : Data_(), Nrows_(0), Ncols_(0)
  {}


  template <class T>
  SparseArray2D<T>::SparseArray2D(const unsigned nrows, const unsigned ncols)
    : Data_(), Nrows_(nrows), Ncols_(ncols)
  {
    if (nrows == 0 || ncols == 0)
      BEXCEPTION("invalid size in Arra2D "<<nrows<<"x"<<ncols);
  }


  template <class T>
  SparseArray2D<T>::SparseArray2D(const SparseArray2D<T>& m)
    : Data_(), Nrows_(0), Ncols_(0)
  { *this = m; }


  template <class T>
  SparseArray2D<T>::~SparseArray2D()
  { /*std::cout <<"destructor ~SparseArray2D()\n"; */ clear(); }


  template<class T>
  SparseArray2D<T>& SparseArray2D<T>::operator=(const SparseArray2D<T>& m)
  {
    // this indirection is necessary to be able to cope with self assignement
    std::map<long long unsigned, T> tmp = m.Data_;
    Data_ = tmp;
    Nrows_ = m.Nrows_;
    Ncols_ = m.Ncols_;
    return *this;
  }


  template<class T>
  void SparseArray2D<T>::clear()
  { Data_.clear(); Ncols_ = Nrows_ = 0; }


  template<class T>
  void SparseArray2D<T>::resize(const unsigned nrows, const unsigned ncols)
  { 
    std::map<long long unsigned, T> tmp;
    unsigned col, row;
    typename std::map<long long unsigned, T>::const_iterator it;
    for (it = Data_.begin(); it!=Data_.end(); it++){
      Map2Indices(it->first, row, col);
      tmp[(long long unsigned)row * (long long unsigned)ncols + 
          (long long unsigned)col] = it->second;
    }
    Data_ = tmp;
    Nrows_ = nrows;
    Ncols_ = ncols;
  }


  template<class T>
  bool SparseArray2D<T>::is_valid(const unsigned row, const unsigned col) const
  {
    if (row >= Nrows_ || col >= Ncols_) return false;
    if (Data_.find(Indices2Map(row, col)) == Data_.end()) return false;
    return true;
  }


  template<class T>
  void SparseArray2D<T>::erase(const unsigned row, const unsigned col)
  {
    BIASASSERT(row < Nrows_ && col < Ncols_);
    typename std::map<long long unsigned, T>::iterator it;
    it = Data_.find(Indices2Map(row, col));
    if (it == Data_.end()) return;
    Data_.erase(it);
  }


  template<class T>
  T& SparseArray2D<T>::operator()(const unsigned row, const unsigned col)
  {
    if (row >= Nrows_ || col >= Ncols_) 
      BEXCEPTION("index out of bounds: array2D size = "<<Ncols_<<"x"<<Nrows_
                 <<"  and index access at ["<<row<<"]["<<col<<"]");
    return Data_[Indices2Map(row, col)];
  }


  template<class T>
  const T& SparseArray2D<T>::operator()(const unsigned row, const unsigned col) const
  {
    if (row >= Nrows_ || col >= Ncols_) 
      BEXCEPTION("index out of bounds: array2D size = "<<Ncols_<<"x"<<Nrows_
                 <<"  and index access at ["<<row<<"]["<<col<<"]");
    typename std::map<long long unsigned, T>::const_iterator it;
    if ((it=Data_.find(Indices2Map(row, col)))==Data_.end())
      BEXCEPTION("matrix is empty at ["<<row<<"]["<<col<<"]");
    return it->second;
  }


} // namespace

#endif
