/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HMatrix.hh"

#include <Math/SVD.hh>




using namespace BIAS;
using namespace std;

HMatrix::HMatrix()
{}


HMatrix::~HMatrix()
{}

void HMatrix::GetMappedRange(const unsigned int OrigWidth, 
                             const unsigned int OrigHeight,
                             BIAS::Vector2<double> & corner,
                             double & sizeWidth, double & sizeHeight) const {
  double minX, maxX, minY, maxY; // extrema
  
  // rect. coordinates of the 4 src. corners.
  BIAS::HomgPoint2D mapped[4];
  
  // 0  3
  // 1  2
  mapped[0]= (*this) * BIAS::HomgPoint2D(0,         0,          1); 
  // H * '0' left,  upper
  
  mapped[1]= (*this) * BIAS::HomgPoint2D(0,         OrigHeight, 1); 
  // H * '1' left,  lower

  mapped[2]= (*this) * BIAS::HomgPoint2D(OrigWidth, OrigHeight, 1); 
  // H * '2' right, lower

  mapped[3]= (*this) * BIAS::HomgPoint2D(OrigWidth, 0,          1); 
  // H * '3' right, upper
      
  // calc. extrema(xmy) of mapped corners
  // init with first point:
  mapped[0].Homogenize(); // let w==1
  minX=maxX=mapped[0][0]; // x euclid
  minY=maxY=mapped[0][1]; // y suclid  
  // homogenize and calc extrema
  for (unsigned int i=1; i<4; i++) {
    mapped[i].Homogenize();
    // find new extrema
    minX = (mapped[i][0]<minX) ? mapped[i][0] : minX; // x
    maxX = (mapped[i][0]>maxX) ? mapped[i][0] : maxX;    
    minY = (mapped[i][1]<minY) ? mapped[i][1] : minY; // y
    maxY = (mapped[i][1]>maxY) ? mapped[i][1] : maxY;
  };
  
  corner=BIAS::Vector2<double>(minX, minY);
  sizeWidth=maxX-minX;
  sizeHeight=maxY-minY;  
}


double HMatrix::GetMappedError(const HomgPoint2D &PointPicOne,
                               const HomgPoint2D &PointPicTwo) const {
  
  HomgPoint2D x1 = PointPicOne;
  HomgPoint2D x2 = PointPicTwo;
  x1.Homogenize();
  x2.Homogenize();
  HomgPoint2D mapped((*this) * x1);
  mapped.Homogenize();
  return (mapped[0]-x2[0])*(mapped[0]-x2[0]) +
    (mapped[1]-x2[1])*(mapped[1]-x2[1]);
}

void HMatrix::GetJacobian(const HomgPoint2D& x, Matrix<double>& Jac) const {
#ifdef BIAS_DEBUG
  if (!x.IsHomogenized()) {
    BIASERR("supplied unhomogenized point !");
    BIASABORT;
  }
  if ((Jac.GetRows()!=2) ||(Jac.GetCols()!=2)) {
    BIASERR("Jacobian has wrong size !");
    BIASABORT;
  }
#endif

  // ( x' y' )^T = H( (x y)^T), assume w==w'==1
  const double denominv = 1.0 / ((*this)[2][0]*x[0]+(*this)[2][1]*x[1]+
    (*this)[2][2]);

  const double denomsquaredinv = denominv * denominv;
  const double enumx = ((*this)[0][0]*x[0]+(*this)[0][1]*x[1]+
    (*this)[0][2]) * denomsquaredinv;
  const double enumy = ((*this)[1][0]*x[0]+(*this)[1][1]*x[1]+
    (*this)[1][2]) * denomsquaredinv;
   
  // x'/dx
  Jac[0][0] = ((*this)[0][0]*denominv - (*this)[2][0]*enumx);
  // x'/dy
  Jac[0][1] = ((*this)[0][1]*denominv - (*this)[2][1]*enumx);
  // y'/dx                                        
  Jac[1][0] = ((*this)[1][0]*denominv - (*this)[2][0]*enumy);
  // y'/dy
  Jac[1][1] = ((*this)[1][1]*denominv - (*this)[2][1]*enumy);
}


int HMatrix::FactorizeAffineMatrixRLeft(const Matrix2x2<double>& A,
                                        Matrix2x2<double>& phi,
                                        Matrix2x2<double>& theta,
                                        double& d1, double& d2){
  // TODO: use svd2x2 here !
  SVD<double> mysvd(A);
  d1 = mysvd.GetS()[0];
  d2 = mysvd.GetS()[1];
  phi = mysvd.GetV().Transpose();
  theta =  mysvd.GetU() * phi;
  // theta[0][0] = theta[1][1] = 0.5 * theta[0][0] + 0.5 * theta[1][1];
  return 0;
}

int HMatrix::FactorizeAffineMatrixRRight(const Matrix2x2<double>& A,
                                         Matrix2x2<double>& phi,
                                         Matrix2x2<double>& theta,
                                         double& d1, double& d2){
  // TODO: use svd2x2 here !
  SVD<double> mysvd(A);
  
  d1 = mysvd.GetS()[0];
  d2 = mysvd.GetS()[1];
  phi = mysvd.GetU().Transpose();
  theta =  mysvd.GetU() * mysvd.GetV().Transpose();
  // theta[0][0] = theta[1][1] = 0.5 * theta[0][0] + 0.5 * theta[1][1];
  return 0;
}

int HMatrix::ComposeAffineMatrixRLeft(Matrix2x2<double>& A,
                                      const Matrix2x2<double>& phi,
                                      const Matrix2x2<double>& theta,
                                      const double& d1, const double& d2){
  int res = 0;
  Matrix2x2<double> d(MatrixZero);
  d[0][0] = d1;
  d[1][1] = d2;
  
  A = theta * phi.Invert() * d * phi;
  return res;
}

int HMatrix::ComposeAffineMatrixRRight(Matrix2x2<double>& A,
                                       const Matrix2x2<double>& phi,
                                       const Matrix2x2<double>& theta,
                                       const double& d1, const double& d2){
  int res = 0;
  Matrix2x2<double> d(MatrixZero);
  d[0][0] = d1;
  d[1][1] = d2;
  A = phi.Invert() * d * phi * theta;
  return res;
}
