/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HMatrixEstimation.hh"

#include <Math//SVD.hh>
#include <Geometry/KMatrix.hh>

using namespace BIAS;
using namespace std;



HMatrixEstimation *pCurrentObjectH = NULL;

bool HMatrixToParams( const HMatrix &Hin,
                                       vector<double> &Params )
{
  if ( Hin.IsZero() )
    {
      BIASERR("A parametrization of a null matrix doesn't make sense!");
      return false;
    }
  HMatrix H=Hin;
  Params.clear();

  // make sure that 3,3 element is positive
  if ( H[2][2] < 0 )
    H.Scale( -1, H);

  // compute the Frobenius norm
  double  Hnorm = 
    H[0][0] * H[0][0] + H[0][1] * H[0][1] + H[0][2] * H[0][2] +
    H[1][0] * H[1][0] + H[1][1] * H[1][1] + H[1][2] * H[1][2] +
    H[2][0] * H[2][0] + H[2][1] * H[2][1] + H[2][2] * H[2][2];
  Hnorm = sqrt(Hnorm);

  Params.push_back(H[0][0]/Hnorm);
  Params.push_back(H[0][1]/Hnorm);
  Params.push_back(H[0][2]/Hnorm);
  Params.push_back(H[1][0]/Hnorm);
  Params.push_back(H[1][1]/Hnorm);
  Params.push_back(H[1][2]/Hnorm);
  Params.push_back(H[2][0]/Hnorm);
  Params.push_back(H[2][1]/Hnorm);

  return true;
}

  /** composes an H-matrix from the seven given parameters
        @return true on success, false on error
        @author mdunda 09 2003 */  
bool ParamsToHMatrix( HMatrix &Mat,
                      const vector<double> &Params )
{
  if ( Params.size() != 8 )
    {
      BIASERR("There must be 8 parameters to compose the H matrix!");
      return false;
    }

  Mat[0][0]=Params[0];
  Mat[0][1]=Params[1];
  Mat[0][2]=Params[2];
  Mat[1][0]=Params[3];
  Mat[1][1]=Params[4];
  Mat[1][2]=Params[5];
  Mat[2][0]=Params[6];
  Mat[2][1]=Params[7];

  /* compute the 3,3 element, using the property from parametrization
     where ||H|| = 1 (||.|| = Frobenius norm) */
  Mat[2][2]=sqrt(1 - 
    (Params[0] * Params[0] +
     Params[1] * Params[1] +
     Params[2] * Params[2] +
     Params[3] * Params[3] +
     Params[4] * Params[4] +
     Params[5] * Params[5] +
     Params[6] * Params[6] +
     Params[7] * Params[7] ));
  return true;
}


void HErrorFunction(int *m, int *n, double *x, 
                    double *fvec, int *iflag) {
  vector<double> params;
  params.reserve(8);
  for (unsigned int i=0; i<8; i++) params.push_back(x[i]);
  BIAS::HMatrix H;
  ParamsToHMatrix(H, params);
   
  HMatrix Hinv;
  BIASASSERT(H.GetInverse(Hinv));
  
  const unsigned int uiDataSize = 4*pCurrentObjectH->p1_->size();
  const vector<BIAS::HomgPoint2D> &p1 = *pCurrentObjectH->p1_;
  const vector<BIAS::HomgPoint2D> &p2 = *pCurrentObjectH->p2_;

  double errorsum = 0;
  for (unsigned int j=0; j<uiDataSize; j+=4) {
    HomgPoint2D ep1=Hinv*p2[j/4];
    ep1.Homogenize();
    fvec[j]   = ep1[0]-p1[j/4][0];
    fvec[j+1] = ep1[1]-p1[j/4][1];
    HomgPoint2D ep2=H*p1[j/4];
    ep2.Homogenize();
    fvec[j+2] = ep2[0]-p2[j/4][0];
    fvec[j+3] = ep2[1]-p2[j/4][1];
    errorsum += (ep1-p1[j/4]).NormL2() +(ep2-p2[j/4]).NormL2(); 
  }
  cout<<errorsum<<" "<<H<<flush<<endl;
}

// int HMatrixEstimation::Optimize(BIAS::HMatrix& H,
//                                 const std::vector<BIAS::HomgPoint2D> &p1, 
//                                 const std::vector<BIAS::HomgPoint2D> &p2) {
// #ifdef BIAS_DEBUG
//   if (p1.size()!=p2.size()) {
//     BIASERR("Vectors of corresponding points have different sizes !!!");
//     BIASABORT;
//   }
//   if (p1.size()<5) {
//     BIASERR("Too few points to optimize: "<<p1.size());
//     return -1;
//   }
// #endif


//   // we need to call an external function from levenberg-marquardt,
//   // prepare data structs
//   // warning: this section is not thread safe !
//   p1_ = &p1;
//   p2_ = &p2;
//   pCurrentObjectH = this;

//   vector<double> vecHParamsStart, vecHParamsResult; 
//   Vector<double> HParamsStart(8), HParamsResult(8);
//   HMatrixToParams(H, vecHParamsStart);
//   for (unsigned int i=0; i<8; i++)  HParamsStart[i] = vecHParamsStart[i];
//   int res;
//   if ((res = LevenbergMarquardt(&HErrorFunction, p1.size()*4,8, 
//                                 HParamsStart, HParamsResult, 
//                                 MINPACK_DEF_TOLERANCE, LM_DEF_MAX_ITER))!=0) {
//     BIASERR("Error in Levenberg-Marquardt-Optimization of H: "<<res);
//     abort();
//     return res;
//   }
//   vecHParamsResult.resize(8);
//   for (unsigned int i=0; i<7; i++)  vecHParamsResult[i] = HParamsResult[i];
//   ParamsToHMatrix(H, vecHParamsResult);
// #ifdef BIAS_DEBUG
//   pCurrentObjectH = NULL;
//   p1_ = NULL;
//   p2_ = NULL;
// #endif

//   return 0;
// }







int HMatrixEstimation::Compute(const vector<BIAS::HomgPoint2D>& points1, 
                               const vector<BIAS::HomgPoint2D>& points2,
                               HMatrix& H) {
 
  int result = 0;
  BIASASSERT(points1.size()==points2.size());
  if ((int) points1.size()<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }
  pointsFrom_=points1;
  pointsTo_=points2;

  const std::vector<BIAS::HomgPoint2D> *pp1 = &points1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &points2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (normalizeHartley_) {
    normalization_.Hartley(points1, p1norm, K1);
    normalization_.Hartley(points2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  Matrix<double> A(2*points1.size(), 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*(*pp1).size(); p+=2, p2++){
 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = (*pp2)[p2][2] * (*pp1)[p2][0];
    A[p+1][1] = (*pp2)[p2][2] * (*pp1)[p2][1];
    A[p+1][2] = (*pp2)[p2][2] * (*pp1)[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = (*pp2)[p2][1] * (*pp1)[p2][0];
    A[p][7] = (*pp2)[p2][1] * (*pp1)[p2][1];
    A[p][8] = (*pp2)[p2][1] * (*pp1)[p2][2];
    
    A[p+1][6] = - (*pp2)[p2][0] * (*pp1)[p2][0];
    A[p+1][7] = - (*pp2)[p2][0] * (*pp1)[p2][1];
    A[p+1][8] = - (*pp2)[p2][0] * (*pp1)[p2][2];
  }
  SVD<double> svd(A);
  //  result = General_singular_value_decomposition(A, S, U, VT);
  svd.GetV(U); 
  VT=U.Transpose();

  for (p=0; p<9; p++){
    H.GetData()[p] = VT[8][p];
  }  

  // undo normalization
  if (normalizeHartley_) H = K2.Invert() * H * K1;

  H_=H;
  return result;
}

int HMatrixEstimation::ComputeAffine(const vector<BIAS::HomgPoint2D>& points1,
                                     const vector<BIAS::HomgPoint2D>& points2,
                                     HMatrix& H) {
  
  // use direct linear transformation on point pairs (p', p^T)
  // p^T crossproduct  HMatrix * p' = 0
  // leads to  A * hvector = 0
  // where A is a matrix compute from p and p'

  BIASASSERT(points1.size()==points2.size());
  
  pointsFrom_=points1;
  pointsTo_=points2;

  // perform hartley normalization, if desired
  const std::vector<BIAS::HomgPoint2D> *pp1 = &points1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &points2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (normalizeHartley_) {
    normalization_.Hartley(points1, p1norm, K1);
    normalization_.Hartley(points2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  // init model matrix with zero
  Matrix<double> A(2*points1.size(), 7, MatrixZero);
  H.SetIdentity();

  // fill rest of model matrix, same as for general H case, but leaving 
  // out last row of H (only 7 parameters of which one is redundant)  
  const unsigned int nummeas = 2*(*pp1).size();
  for (register unsigned int p=0, p2=0; p<nummeas; p+=2, p2++){
    A[p][3]   = -(A[p+1][0] = (*pp2)[p2][2] * (*pp1)[p2][0]);
    A[p][4]   = -(A[p+1][1] = (*pp2)[p2][2] * (*pp1)[p2][1]);
    A[p][5]   = -(A[p+1][2] = (*pp2)[p2][2] * (*pp1)[p2][2]);
    
    A[p][6]   =   (*pp2)[p2][1] * (*pp1)[p2][2];
    A[p+1][6] = - (*pp2)[p2][0] * (*pp1)[p2][2];
  }
  
  // solve using svd (at least 1d solution space)
  Vector<double> S;
  Matrix<double> U, VT; 
  SVD<double> svd(A);
  svd.GetV(U);
  VT=U.Transpose();
  const double  normalization = 1.0 / VT[6][6];
  double* pHData = H.GetData();
  for (int p=0; p<6; p++){
    *pHData++ = VT[6][p] * normalization;
  }    
  
  // undo normalization
  if (normalizeHartley_) H = K2.Invert() * H * K1;

  H_=H;

  return 0;
}

int 
HMatrixEstimation::Compute(const std::vector<Vector2<double> >& frompoints, 
                           const std::vector<Vector2<double> >& topoints, 
                           HMatrix &H)
{
#ifdef BIAS_DEBUG
  if (frompoints.size()!= topoints.size()) {
    BIASERR("lists fromPoints and toPoints are of different size!");
    return -5;
  }
#endif
  vector<BIAS::HomgPoint2D> fromPointsHomg( frompoints.size() );
  vector<BIAS::HomgPoint2D> toPointsHomg( topoints.size() );
  for (unsigned int i=0;i<frompoints.size();i++) {
    fromPointsHomg[i][0]= frompoints[i][0];
    fromPointsHomg[i][1]= frompoints[i][1];
    fromPointsHomg[i][2]= 1.0;
    toPointsHomg[i][0]= topoints[i][0];
    toPointsHomg[i][1]= topoints[i][1];
    toPointsHomg[i][2]= 1.0;    
  }
  return Compute(fromPointsHomg, toPointsHomg, H);
}

double HMatrixEstimation::GetError()
{
  double err=0;
  BIAS::HomgPoint2D mapped;
  for (unsigned int c=0;c<pointsFrom_.size();c++) {
    mapped= H_ * pointsFrom_[c];
    mapped.Homogenize();
    mapped-=pointsTo_[c];
    err+=mapped[0]*mapped[0] + mapped[1]*mapped[1];
  }
  err/= (double)pointsFrom_.size();
  return err;
}


