/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// define it for extra debugging output (JW)
//#define ImageIO_DEBUG_IO

#include "ImageIO.hh"
#include <Common/FileHandling.hh>

#include <Debug/DebugSimple.hh>

#include <fstream>
#include <sstream>

#include "ImageConvert.hh"

// for transfrom:
#include <algorithm>
#include <cctype>

// for O_APPEND
#include <fcntl.h>

// for perror
#include <stdio.h>

using namespace BIAS;
using namespace std;

char *ImageIO::MemAlignedBuffer_ = NULL;
unsigned long int ImageIO::MemAlignedBufferSize_ = 0;



/// helper function for ExtensionName type casting @author Jan Woetzel
char myTolower(char c) {
  return char(tolower(c));
}



// JW try to Read MIP image and if that fails try all import functions. 
int ImageIO::Load(const string& FileNameConst, ImageBase &img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  string FileName; // (FileNameConst); // non const copy for tilde expansion
  BIAS::FilenameExpansion(FileNameConst, FileName); // tilde

  int result = 0;
  //ifstream is(FileName.c_str());
  ifstream is;
  is.open( FileName.c_str() );
  char identifier[IDENTIFIER_LENGTH];

  if (!is || !is.good()){
    // file does not exist/is not readable!
#if defined(BIAS_DEBUG) || defined(ImageIO_DEBUG_IO)
    BIASWARN("could not open fstream for file "<<FileName<<" with cwd="<<BIAS::GetCwd());
#endif
    result = -1;
  } else {
    is.read(identifier, IDENTIFIER_LENGTH);
    if (!is.good()){
      result = -2;
      is.close();
    } else {
      if (identifier[0] == 'M' && identifier[1] == 'I' && 
        identifier[2] == 'P'){ // mipimage
          is.close();
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
          COUT("MIP format detected\n");
#endif

          result = Read(FileName, img);
      } else { 
        // foreign image format
        is.close();
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
        COUT("image not in mip format\n");
#endif
        result = ImageIO::ImportImage(FileName, img);
      }
    }
  }
  return result;
}





// JW: major rewrite 2005
int ImageIO::ImportImage(const string& filename, ImageBase &result ) 
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res=-1; // init as file not read
  ifstream ifs;
  unsigned char tmpline[512];

#ifdef   WIN32
  ifs.open(filename.c_str(),ios::binary);
#else  //not WIN32
  ifs.open(filename.c_str());
#endif //WIN32

  if (!ifs) {
    BIASERR("ImportImage(): open of file \""<<filename<<"\" failed");
    perror(filename.c_str());
    ifs.close();
    return -1;
  }


  // read first two chars to get a hint what this file is
  ifs >>tmpline[0]>>tmpline[1];
  if (!ifs){
    res=-1;
    BIASERR("unable to read from "<<filename);
    ifs.close();
    return res;
  }

  // Check for famous P5/P6 in PNM, PGM, PBM to determine the fileformat
  // and call our own reader if appropriate.
  if (tmpline[0] =='P' && tmpline[1] == '5') {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("reading single channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 1); 
    if (res!=BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P5 format detected.");
    result.SetColorModel(ImageBase::CM_Grey);
  } else if (tmpline[0] =='P' && tmpline[1] == '6') {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("reading 3 channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 3); 
    if (res!= BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P6 format detected.");
    result.SetColorModel(ImageBase::CM_RGB);
  } 

  // other loaders do not use file stream but operate on filename directly 
  // and open theirselves.
  ifs.close();
  
  return res;
} // ImportImage






int ImageIO::ImportImagePnm_(ifstream& ifs, ImageBase &result, int depth)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int xres=-1, yres=-1, maxval=-1, read=0;
  int res=0;
  char c;

  if (!result.IsEmpty())
    result.Release();
  // start from the beginning and read line with P5 or P6
  ifs.seekg(0,ios::beg);

  ifs >> c; 
  if (c!='P'){
    BIASERR("invalid magic cookie");
    return -1;
  }
  ifs >> c; 
  if (c!='5' && c!='6'){
    BIASERR("invalid magic cookie");
    return -1;
  }

  // this allows an arbitary amount of whitespace in between the header
  while (read<3){
    ifs >> c; 
    ifs.putback(c);
    if (c=='#'){
      ifs >> *result.GetMetaData();
      if (!ifs){
        BIASERR("error reading meta data");
        return -1;
      }
    } else if (c>29 || c<40){ // number 0-9
      switch (read){
    case 0:
      ifs >> xres;
      //cerr << "xres: "<<xres<<endl;
      read++;
      break;
    case 1:
      ifs >> yres;
      //cerr << "yres: "<<yres<<endl;
      read++;
      break;
    case 2:
      ifs >> maxval;
      //cerr << "maxval: "<<maxval<<endl;
      read++;
      break;
    default:
      BIASERR("funny error: "
        <<"there should be only 3 numerical values in pgm header");
      return -1;
      break;
      }; // switch
    } else { // if (c>29 || c<40){
      BIASERR("corrupted image header "<<c);
    }
  }
  ifs.read(&c, 1*sizeof(char)); // last whitespace

  //cerr <<"image has size "<<xres<<"x"<<yres<<"  with maxval "<<maxval<<endl;
  if (maxval==255){
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedchar &&
      result.GetStorageType()!=ImageBase::ST_invalid){
        BIASERR("warning trying to load an unsigned char image into an image of"
          <<" storage type "<<result.GetStorageType());
        //BIASASSERT(false);
        BIASABORT;
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedchar);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), xres * yres * depth);
  } else if (maxval==4095){ // 12 bit images
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedshortint &&
      result.GetStorageType()!=ImageBase::ST_invalid){
        BIASERR("warning trying to load an unsigned short int image into an "
          <<"image of storage type "<<result.GetStorageType());
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedshortint);  
    result.SetBitDepth(12);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), 
      sizeof(unsigned short int) * xres * yres * depth);
    // check if bytes need swapping
    unsigned short int *p=(unsigned short int *)result.GetImageData(), *pe;
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    // check if pixel > maxval exist
    bool swap=false;
    while (p<pe) { swap = (swap || (*p > maxval)); p++; }
    // swap bytes if pixel > maxval existed
    if (swap){
      BIASERR("maxval > "<<maxval<<" : swapping LSB and MSB");
      char *cp=(char *)result.GetImageData(), *cpe;
      cpe=(char *)((unsigned short int *)cp + result.GetPixelCount()*
        result.GetChannelCount());
      char c;
      while (cp<cpe){
        c=cp[0]; cp[0]=cp[1]; cp[1]=c;
        cp+=2;
      }
    }
#ifdef BIAS_DAIMLERCHRYSLER
    // check image, since bosch fills last line with values above 4095
    p=((unsigned short int *)result.GetImageData());
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    while (p<pe) {
      if (*p>maxval) *p=maxval; 
      p++; 
    }
#endif
  }

  if (!ifs.good()){
    BIASERR("error reading from ifs");
    res=-1;
  }
  return res;
}


int ImageIO::Save(const std::string& FileName,
                  const  ImageBase &img,
                  const enum TFileFormat FileFormat,
                  const bool sync,
                  const int jpeg_quality,
                  const bool forceNewID,
                  const bool & writeMetaData)
{
  return ExportImage(FileName, 
    img,
    FileFormat, 
    sync,
    jpeg_quality, 
    forceNewID,
    writeMetaData );
}



int ImageIO::ExportImage(const string& c_filename, 
                         const ImageBase &img_const,
                         const enum TFileFormat c_FileFormat, 
                         const bool sync,
                         const int c_jpeg_quality, 
                         const bool forceNewID,
                         const bool & writeMetaData) 
{ 
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  bool TriedIt = false;

  ImageBase img(img_const);
  // do NOT scale or shift in this routine. 
  // Do it before calling this routine if you need it. (JW 04/2003)

  int result = -1;
  int quality = c_jpeg_quality;

  // non const to change for auto detection: 
  TFileFormat FileFormat=c_FileFormat;
  if (FileFormat==FF_auto)
  {
    // determine by filename extension: 
    FileFormat=GetFileFormat(c_filename);
    // if we couldn't determine BIAS FileFormat 
    // give the external libs their own "low level" try
    if (FileFormat==FF_unknown)
      FileFormat=FF_auto;
  };

  // non const for adding suffix
  string filename(c_filename);
  BIAS::FilenameExpansion (c_filename, filename); // tilde


  if (img.GetStorageType()==ImageBase::ST_invalid){
    BIASERR("cannot save images with StoragType ST_invalid. ");
    return -1;
  }


 

  // -------------------------------------------------------------
  // ---  try to write with our internal format handlers, first.
  // -------------------------------------------------------------

  if (result!=BIAS_IO_SUCCESS) {
    // construct valid filename with appropriate extension
    switch (FileFormat) 
    {
    case FF_mip:
      filename =ImageIO::ExtensionName( filename, GetExtension(FileFormat) );
      result = Write(filename, img, sync, forceNewID);
      TriedIt = true;
      break;
    case FF_pgm:
    case FF_ppm:
      {
        filename = ExtensionName( filename, GetExtension(FileFormat) );
        // use our internal impl. only for 8bit, 16 bit is impl. incorrectly
        if (img.GetDepth()==1) /* Byte per Channel */
        {
          if ((FileFormat==FF_pgm) && (img.GetChannelCount()!=1)) 
          {
            // we need to convert: 
            Image<unsigned char> greyIMG(img);
            ImageConvert::IP_ToGrey(greyIMG);
            result = ExportImagePnm_(filename,greyIMG,sync,writeMetaData);
            TriedIt = true;
          } else {
            // no slow conversion, is already correct fmt 
            result = ExportImagePnm_(filename,img,sync,writeMetaData);
            TriedIt = true;
          };
        }
      }
      break; 
    case FF_jpg:
      filename = ImageIO::ExtensionName( filename, GetExtension(FileFormat) );
      if (c_jpeg_quality < 0 || c_jpeg_quality > 100) 
      {
        BIASERR("The quality "<<c_jpeg_quality
          <<" is out of valid range 0..100 % falling back to 100% full quality." );
        quality=100;
      }
      break;
    case FF_raw4096:
      BIASERR("writing raw4096 is not (yet) implemented in BIAS, only on WIN32 scorpion grab application.");
      return -7;
      break;

      //  // we could add our internal handlers for pnm, here.
      //  // however - the IMageMagicj and DevIL versions 
      //  // are more reliable, in particular for hdr formats. 

    case FF_auto: 
    case FF_unknown:
      // leave the filename as it is for fmt detection by filename.
      break;
    case FF_bmp:
    case FF_pbm:
    case FF_tif:
    case FF_png:
    case FF_viff:
    case FF_exr:
    case FF_dds:
      filename = ExtensionName( filename, GetExtension(FileFormat) );
      break;
    default:
      BIASERR("Unknown image file format for saving "<<c_filename
        <<"Nothing saved!"
        );
      return -1;
    } // switch
  };

  // --------------------------------------------------
  // ---   Try to write using external libraries.   ---
  // --------------------------------------------------

  if (!TriedIt) 
    BIASERR("Image not even tried to be written");

  // final result:
  return result;
}



int ImageIO::ExportImagePnm_(const string &FileName, 
                             const ImageBase &img, 
                             bool sync,
                             const bool & writeMetaData)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  ofstream ofs;
  // bool sync=false;

#ifdef   WIN32
  ofs.open(FileName.c_str(),ios::binary);
#else  //not WIN32
  ofs.open(FileName.c_str());
#endif //WIN32

  if (!ofs) {
    BIASERR("ExportImagePnm_() : open of file "<<FileName<<" failed");
    perror(FileName.c_str());
    return -1;
  }

#ifdef BIASDEBUG
  if (img.GetDepth()!=1)
    BIASERR(FUNCNAME <<" supports only 8 bit per Channel images. Please call "
    <<"ExportImage or Save instead.");
#endif // BIASDEBUG

  switch (img.GetChannelCount()) {
  case 3: // 8bit ppm
    ofs <<"P6"<<endl;
    break;

  case 1: // 8bit pgm
    ofs << "P5"<<endl;
    break;
  default:
    BIASERR("ExportImagePnm_() : unsupported number of channels: "<<
      img.GetChannelCount());
    return -1;
  }

  if (writeMetaData){
    img.GetMetaData()->WriteAscii(ofs);
    if (!ofs){
      BIASERR("writing MetaData failed");
      return -2;
    } 
  };

  ofs << img.GetWidth()<<" "<<img.GetHeight()<<endl;
  ofs << "255"<<endl;

  ofs.write(reinterpret_cast<const char *>(img.GetImageData()), img.GetSizeByte());
  if(!ofs) {
    BIASERR("ExportImagePnm_() : write failed");
    return -1;
  }

  if (sync) {
    // flush streambuf from c++
    ofs.flush();
    // flush system buffers
#ifndef _WIN32
#  if __GNUC__ < 3
    fsync(ofs.rdbuf()->fd());
#  else
    ofs.rdbuf()->pubsync();
#  endif
#else // _WIN32
    ofs.rdbuf()->pubsync();
#endif //_WIN32
  }

  ofs.close();
  return 0;
}


std::string ImageIO::ExtensionName(const std::string & oldName, 
                                   const std::string & newExtension) 
{
  string name    = oldName; // full name
  string ext     = BIAS::Extension( oldName );
  string newExtLC= BIAS::LowerCase( newExtension );

  // catch empty file name: 
  if (name.size()<1)
    name="out___"; // default

  // append extension if it is not the correct case insensitive extension: 
  if (ext!=newExtLC) {
    BIASASSERT(name.size()>0);
    const char c = name[name.size()-1];
    if (c!='.')
      name+= ".";
    name+=newExtLC;
  }
  return name;
}


int ImageIO::Write(const string& FileName, const ImageBase &img_const, 
                   bool sync, bool forceNewID) 
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res = 0;
  ofstream os;

  ImageBase img(img_const);

//   // assign new id if forced or there is none yet
//   if ((forceNewID) || (!img.GetUID().IsValid())) {
//     img.SetUID(UUID::GenerateUUID());
//   }

  // avoid ".mip.mip" (JW) 
  string file = ImageIO::ExtensionName( FileName, "mip");

#ifdef   WIN32
  os.open(file.c_str(),ios::binary);
#else  //not WIN32
  os.open(file.c_str());
#endif //WIN32

  if (!os){
    string Errorstring="Error opening "+file+" writing";
    perror(Errorstring.c_str());
    return -1;
  }

  os << img;

  res = (os.good())?(0):(-5);

  if (res==0){
    //cerr <<"Writing MetaData\n";
    //img.GetMetaData()->Dump();
    img.GetMetaData()->WriteBinary(os);
    if (!os){
      BIASERR("writing MetaData failed");
      res=-6;
    }
  }

  if (sync) {
    // flush streambuf from c++
    os.flush();
    // flush system buffers
#ifndef _WIN32
#  if __GNUC__ < 3 // in gcc 2.9x we can get a POSIX file descriptor 
    fsync(os.rdbuf()->fd());
#  else   // in gcc 3.x this is prohibited :-(
    os.rdbuf()->pubsync(); 
    //     ::sync(); // this is systemcall sync() to sync all buffers    
#  endif
#else // _WIN32
    /** @todo Sync written files and buffers to disk under WIN32 
    (evers 2004-10-14) */
    // 	os.rdbuf()->pubsync();
    //     ::sync();
#endif // _WIN32
  }
  os.close();

#ifndef _WIN32 // alternative way to sync, reopen file to get a handle
  int fd = open(file.c_str(),O_APPEND);
  fsync(fd); // and now fsync()
  close(fd);
#endif 
  return res;
}


int ImageIO::Read(const string& FileName, ImageBase &img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res=0;
  ifstream is;

#ifdef   WIN32
  is.open(FileName.c_str(),ios::binary);
#else  //not WIN32
  is.open(FileName.c_str());
#endif //WIN32

  if (!is){
    string Errorstring="Error opening "+FileName+" for reading";
    perror(Errorstring.c_str());
    return -1;
  }
  is >> img;


  res = (is.good())?(0):(-5);

  if (res==0){
    //cerr << "ImageIO::Read() : "<<is.tellg()<<endl;
    is >> *img.GetMetaData();
    if (!is && !is.eof()){
      BIASERR("error reading MetaData");
      res = -6;
    }
  }

  //img.GetMetaData()->Dump();

  is.close();


  return res;
}




// JW
int ImageIO::ImportRAWwithHeader(const std::string & filename, 
                                 BIAS::ImageBase & img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  // header data size:
  unsigned int hdrSZ=0;
  // tags to find
  unsigned int Cols=0;
  unsigned int Rows=0;


  { // scope
    ifstream is;
    is.open(filename.c_str()); 
    //is.open(filename.c_str(), ios::binary);
    //is.open(filename.c_str(), ios_base::in |ios_base::binary );
    if (!is){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
      BIASERR("could not open image file: "<<filename);
#endif
      is.close();
      return -1;
    }
    // OK
    string line;
    getline(is, line);
    //cout<<"DBG "<<FUNCNAME<<"line: "<<line<<endl;
    if (line.size()<3) return -2; // too small. 

    // parse header: 
    string ref7("RAW4096"); // 7 Bytes instead of std. 8 for backward compatibility 
    string ref("RAW04096");
    if ( (0==line.compare(0, ref.size(), ref))
      || (0==line.compare(0, ref7.size(), ref7)) )
    {
      COUT("RAW04096/RAW4096 found."<<endl);
      hdrSZ=4096;
    } else {
      ref="RAW08192";
      if (0==line.compare(0, ref.size(), ref)){
        hdrSZ=8192;
      } else {
        ref="2048";
        if (0==line.compare(0, ref.size(), ref)){
          hdrSZ=2048;
        } else {
          ref="1024";
          if (0==line.compare(0, ref.size(), ref)){
            hdrSZ=1024;
          } else {
            ref="0512";
            if (0==line.compare(0, ref.size(), ref)){
              hdrSZ=0512;
            } else {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
              BIASERR(filename<<" is no RAW4096/RAWnnnn image. aborting.");
#endif
              return -3;
            }
          }
        }
      }
    }


    // create string that can be used as char array:
    string hdr(hdrSZ,' ');
    is.seekg(ios::beg);
    // read data as a block: 
    is.read(&(hdr[0]), hdrSZ);

    bool ok=true;
    //ok = ok && (0==Parse(hdr, "Cols=", Cols));
    ok = ok && (0==ParseRAWwithHeader(hdr, "RowInc=", Cols));
    ok = ok && (0==ParseRAWwithHeader(hdr, "Rows=", Rows));

    if (ok) cout<<"found dim "<<Cols<<"x"<<Rows<<endl;
    is.close();
  }

  if (!img.IsEmpty()) img.Release();
  // HACK: content of RAW is fixed for scorpion 1600x1200 
  // as RAW 1 plane Bayer pattern, 8 bit. (JW)  
  img.Init(Cols, Rows, 1);
  img.SetColorModel(BIAS::ImageBase::CM_Bayer_RGGB);

  //
  // reopen file with C fopen intead of C++ fstream
  // USE C because of higher performance (measured 15%)
  //
  //TODO 
  // FIXME: Linux reads only one line - thu sit is not using binary mode! (jw)
  FILE * fp=NULL;
  //fp = fopen(filename.c_str(), "rb");
  fp = fopen(filename.c_str(), "rt");
  if (fp==NULL){
    cout<<"could not open "<<filename<<" for C reading"<<endl;
    return -3;
  }
  //OK
  unsigned int dataSZ=Cols*Rows;
  //memset(&(img.GetImageData()[0]), 0, dataSZ);
  {
    char * p_data = new char[dataSZ];
    memset(p_data, 0, dataSZ);

    unsigned int readBytes=0;
    // skip header 
    readBytes=fread( p_data, hdrSZ, 1, fp );

    // read data:
    readBytes=fread( p_data, dataSZ, 1, fp );

    img.CopyIn_NoInit(p_data);
    //tmp.CopyIn_NoInit(p_data);

    fclose(fp);
    delete p_data;
  }

  return 0; // OK
}


// JW
int ImageIO::ParseRAWwithHeader(const std::string & buf, 
                                const std::string & tag, 
                                unsigned int & retVal)
{ 
  string::size_type posStart, posEnd;
  string valueStr;
  posStart=buf.find(tag, 0);
  if (posStart!=string::npos){
    // found tag
    posEnd = buf.find( string("\n"), posStart);
    if (posEnd!=string::npos){
      valueStr=buf.substr(posStart+tag.size(), posEnd-posStart-tag.size());
      stringstream ss;
      ss.str(valueStr);
      ss>>retVal;
      cout<<"valueStr="<<valueStr<<" : "<<retVal<<endl;
    }
  } else {
    cout<<"not found!"<<endl;
    return -1;
  }
  // OK.
  return 0;
}






std::string 
ImageIO::GetExtension(const BIAS::ImageIO::TFileFormat & fmt)
{
  string s;
  switch (fmt)
  {
  case FF_mip:
    s="mip"; break;
  case FF_ppm:
    s="ppm"; break;
  case FF_pgm:
    s="pgm"; break;
  case FF_pbm:
    s="pbm"; break;
  case FF_jpg:
    s="jpg"; break;
  case FF_tif:
    s="tif"; break;
  case FF_png:
    s="png"; break;
  case FF_viff:
    s="vif"; break;
  case FF_exr:
    s="exr"; break;
  case FF_dds:
    s="dds"; break;
  case FF_raw4096:
    s="raw"; break;
  case FF_bmp:
    s="bmp"; break;
    break;
  case FF_Real32:
    s="real32"; break;

  case FF_auto:
  case FF_unknown:
  default:
    BIASERR("Unknown FileFormat "<<fmt<<" Please add switch case");
  }
  return s;
}



BIAS::ImageIO::TFileFormat 
ImageIO::GetFileFormat(const std::string &str)
{
  // suffix = "dot"+extension 
  string ext=BIAS::LowerCase( BIAS::Extension(str) );
  // compare ext to all FF_* fixed formats: 
  if (ext=="mip") {
    return FF_mip;
  } else if (
    (ext=="ppm") ||
    (ext=="pnm") /* pNm can be pbm, pgm, ppm. assume color for pnm. */
    )
  {
    return FF_ppm;
  } else if (ext=="pgm") {
    return FF_pgm;
  } else if (ext=="pbm") {
    return FF_pbm;
  } else if (
    (ext=="jpg") ||
    (ext=="jpe") ||
    (ext=="jpeg") 
    )
  {
    return FF_jpg;
  } else if (
    (ext=="tif") ||
    (ext=="tiff") 
    )
  {
    return FF_tif;
  } else if (ext=="png") {
    return FF_png;
  } else if (
    (ext=="vif") || 
    (ext=="vff") ||
    (ext=="viff")) 
  {
    return FF_viff;
  } else if (ext=="exr") {
    return FF_exr;
  } else if (
    (ext=="dds")  ||
    (ext=="dxt")  ||
    (ext=="dxt1") ||
    (ext=="dxt3") ||
    (ext=="dxt5") ||
    (ext=="dxtc") ||
    (ext=="s3tc") 
    )
  {
    // S3's S3TC format and DXTC are essentially the same
    return FF_dds;
  } else if (
    (ext=="raw") ||
    (ext=="raw4096") ||
    (ext=="scorpion") 
    )
  {
    return FF_raw4096;
  } else if ((ext=="real32") || (ext=="f32")){
    // Mathematica Real32 raw JW
    return FF_Real32;
  } else {
    // unknown ext. 
    // give the external imaeg writer libsa try,
    // they may support a wider raneg of extension.
    // Thsu use "FF_auto" on returned "FF_unknown"
    return FF_unknown;
  }
}


//#ifdef BIAS_HAVE_RADIANCE
// JW
int BIAS::ImageIO::ImportRADIANCE(const std::string& FileName, ImageBase& result)
{
  ifstream is;
  is.open(FileName.c_str(), ios::binary);
  if (!is){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not open file "<<FileName);
#endif
    return -1;
  }
  //is.seekg(0,ios::beg);
  string buf;
  unsigned int line=0;
  getline(is, buf); line++;
  //cout<<line<<". = "<<buf<<endl;

  const string signatureRAD("#?RADIANCE");
  if (buf!= signatureRAD )
  {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("signatureRAD "<<signatureRAD<<" not found in file "<<FileName);
#endif
    return -2;
  };
  // OK, is a RADIANCE file. 

  bool flipY=false, flipX=false;
  unsigned int width=0, height=0;
  //float exposure=1.0;

  // parse for FORMAT
  while (is){    
    getline(is, buf); line++;
    //cout<<line<<". = "<<buf<<endl;
    // skip comments: 
    if (0==buf.compare(0,1,"#")) {
      //cout<<"found comment: "<<buf<<endl;
      continue;
    }
    if (0==buf.compare(0,6,"FORMAT")) {
      // check format: 
      if (0==buf.compare("FORMAT=32-bit_rle_rgbe")){
        break; // OK
      } else {
        BIASERR("Only 32-bit_rle_rgbe FORMAT supported but "<<buf<<" found. Failed to load.");
        return -5;
      }
    };
  }

  // parse for -Y or +Y 
  while (is){    
    getline(is, buf); line++;
    //cout<<line<<". = "<<buf<<endl;
    // skip comments: 
    if (0==buf.compare(0,1,"#")) {
      //cout<<"found comment: "<<buf<<endl;
      continue;
    }
    if (0==buf.compare(0,2,"-Y") 
      || 0==buf.compare(0,2,"-Y")) 
    {
      // parse x and y value: 
      stringstream ss(buf);
      string dummy;
      ss>>dummy; // Y
      if (dummy=="+Y"){
        flipY=true;
      } else if (dummy=="-Y") {
        // OK, no flip
      } else {
        BIASERR("expected +Y or -Y but failed.");
        return -8;
      }
      // read Y value
      ss>>height;

      ss>>dummy;// X
      if (dummy=="+X"){
        // OK, no flip
      } else if (dummy=="-X") {
        flipX=true;
      } else {
        BIASERR("expected +X or -X but failed.");
        return -9;
      }
      // read X value
      ss>>width;
      break;
    }
  };
  cout<<FUNCNAME<<" found "<<width<<"x"<<height<<" RADIANCE rle RGBE image. "<<FileName<<endl;
  // get RGBE data:
  if (!result.IsEmpty()) result.Release();
  result.Init(width, height, 4, BIAS::ImageBase::ST_unsignedchar, true );

  // get data
  char col[4]={0};
  unsigned int y=0;
  //unsigned int x=0;
  //int pos=0;
  //int start=0;
  //int step=1;
  for (y=0; y<height; y++) {
    //start = bFlippedY ? (height-y-1)*width : y*width;
    //step = bFlippedX ? -1 : 1;

    // read 4 Bytes of line:
    is.read(col,4);
    // is the image /line RLE compressed? 
    if ( (col[0]==0x2) 
      && (col[1]==0x2) 
      && (col[2]==0x0) )
    {
      BIASERR("unsupported new RADIANCE compression detected in line y="<<y);
      // clear the line: 
      memset( result.GetImageDataArray()[y], 0, 4*width);

    } else if ( (col[0]==0x1) 
      && (col[1]==0x1) 
      && (col[2]==0x1) )
    {
      BIASERR("unsupported old RADIANCE run scheme in line y="<<y);
      memset( result.GetImageDataArray()[y], 0, 4*width);

    } else {

      BIASERR("sorry, RADAINCE not completely implemented TODO: RLE (un-)compression is missing. (jw)");
      return -99;
      /// \todo: decode the runlength encode line, map the RGBE to float or do it in ColorConverion?

      /*
      // OK, assume line contains only raw data
      //start = flipY ? (height-y-1)*width : y*width;
      //step  = flipX ? -1 : 1;
      // Do the run
      int nPixel = ((int) col[3])&0xFF;
      unsigned char r = pBuffer[(pos-step)*4];
      unsigned char g = pBuffer[(pos-step)*4+1];
      unsigned char b = pBuffer[(pos-step)*4+2];
      unsigned char e = pBuffer[(pos-step)*4+3];

      for( int i=0; i<num; i++ ) {
      pBuffer[pos*4] = r;
      pBuffer[pos*4+1] = g;
      pBuffer[pos*4+2] = b;
      pBuffer[pos*4+3] = e;
      pos += step;
      }

      x += num-1;
      */
    }
  }
  return 0; // OK
}
//#endif // BIAS_HAVE_RADIANCE 



// JW
int BIAS::ImageIO::ImportReal32(const std::string& FileName, ImageBase& result)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  ifstream is;
  is.open(FileName.c_str(), ios::binary);
  if (!is){
    BIASERR("Could not open file "<<FileName);
    is.close();
    return -1;
  }

  // header: 4 float
  // format, width, height, nchannels
  vector<float> header(4, 0);
  unsigned int version=0;
  if (is)
    is.read( (char*)(&(header[0])), header.size()*sizeof(float) );

  if (!is)
    return -101;

  // supported format? simple check for now. 
  const float fmt=header[0];
  if ((1.0f <= fmt) && (fmt<=1.01f)) {
    version=1;
  } else {
    return -1; // unsupported
  }
  unsigned int width     = (unsigned int)( header[1] );
  unsigned int height    = (unsigned int)( header[2] );
  unsigned int nchannels = (unsigned int)( header[3] );

  // only small GPU suitable images are written, currently: 
  // check useful header values:
  if (!(0<width)) 
    return -2;
  if (!(width<=4096)) 
    return -3;
  if (!(0<height)) 
    return -4;
  if (!(height<4096)) 
    return -5;
  if (!(0<nchannels)) 
    return -6;
  if (!(nchannels<=4)) 
    return -7;

  // OK

  // (re-)allocate
  if (!result.IsEmpty())
    result.Release();
  result.Init(width, height, nchannels, ImageBase::ST_float, true);

  // read the raw data:
  BIASASSERT(is);
  is.read( (char*)(result.GetImageData()), 
    sizeof(float)*width*height*nchannels );

  is.close();
  return 0;
}



#ifdef WIN32
// implement WINAPI specific calls for BMP loading via disk or memory

int BIAS::ImageIO::Import_BITMAP_winapi(const std::string& FileName, 
                                        ImageBase& result,
                                        const bool flipY )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  // the WINAPI interface
  HBITMAP hbm=NULL;

  // convert from stl string to WINAPI string
  // TODO: what about UNICODE strings: use LPTSTR ?
  LPCSTR szFileName = FileName.c_str();

  // load the image from disk into a DIBSection
  hbm = (HBITMAP)
    LoadImage( NULL, LPCWSTR(szFileName), IMAGE_BITMAP, 0, 0,
    LR_LOADFROMFILE | LR_CREATEDIBSECTION | LR_DEFAULTSIZE );

  // check success
  if( hbm == NULL ){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("could not load "<<FileName );
#endif 
    return -1; // error, could not load
  }

  // we have a valid HBITMAP, now.
  // create object and copy  in, now.
  return BIAS::ImageIO::Import_HBITMAP_winapi(hbm, 
    result,
    flipY );
}


int BIAS::ImageIO::Import_HBITMAP_winapi(const HBITMAP & hbm, 
                                         ImageBase& result,
                                         const bool flipY )
{
  // create object from handle
  BITMAP bm;
  const int nBytesHeader = 
    GetObject( hbm, sizeof(BITMAP), &bm );
  if (nBytesHeader==0) {
    BIASERR("could not GetObject");
    return -1;
  }
  //cout<<"created object with "<<nBytes<<" Bytes."<<endl; // typically 24

  // we have a valid BITMAP, now, 
  // load pixel data
  return BIAS::ImageIO::Import_BITMAP_winapi(bm, 
    result,
    flipY );
}


int BIAS::ImageIO::Import_BITMAP_winapi(const BITMAP & bm, 
                                        ImageBase& result,
                                        const bool flipY )
{
  // If  DIBSection is 256 color or less --> color table
  if( ( bm.bmBitsPixel * bm.bmPlanes ) <= 8 ){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not load winapi BITMAP because it contains a color table (not implemented)" );
#endif 
    return -2;
  }

  // tested only with 3 channel(BGR) interleaved images with one plane
  if (bm.bmPlanes != 1){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not load winapi BITMAP because it contains more than one plane." );
#endif 
    return -3;
  }

  //
  // copy pixel data into BIAS layout
  // assume 3 channel uc (3*8=24bit) interleaved DIB for now.
  //
  {
    BIAS::Image<unsigned char> img(bm.bmWidth, bm.bmHeight, 3 );
    img.SetColorModel(BIAS::ImageBase::CM_BGR); // Windows DIB = BGR

    const int nBytes = bm.bmWidthBytes*bm.bmHeight;
    memcpy(img.GetImageData(), (unsigned char*)bm.bmBits, nBytes );
    if (flipY){
      // Windows BMP has down-up vertical order like OpenGL
      img.Flip();
    }

    // Image<uc> --> ImageBase
    result = img;
  }

  return EXIT_SUCCESS;
}


#endif // WIN32





#ifdef BIAS_HAVE_TIFF
// see http://www.ibm.com/developerworks/linux/library/l-libtiff/   JW
int BIAS::ImageIO::ExportTIFFLIB(const std::string& FileName, 
                                 const ImageBase& img,
                                 const unsigned int rowsPerStripArg,
                                 const unsigned int compressionAlgo,
                                 const string & comment )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  try {

    TIFF* tif = TIFFOpen(FileName.c_str(), "w");
    if (tif==NULL){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
      BIASERR("Could not open image for TIFFLIB for writing:" << FileName);
#endif
      return -1;
    }
    BIASASSERT(tif!=NULL);
    // work:

    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH,      img.GetWidth() );
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH,     img.GetHeight() );
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, img.GetChannelCount() );
    const unsigned short bitspersample = 8* (unsigned short)img.GetDepth(); // img.GetBiTDepth()
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, bitspersample ); // more than enough
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT); // origin 0,0
    // assume planar tiff cfg., needs one channel or interleaved pixel data
    BIASASSERT((img.GetChannelCount()==1) || (img.IsInterleaved()) ); 
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG);


    //TIFFSetField(tif, TIFFTAG_COMPRESSION,     COMPRESSION_NONE); //(300 msec)
    //TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW); // works, lossless Lempel Ziff Welch  (650msec)
    //TIFFSetField(tif, TIFFTAG_COMPRESSION,     COMPRESSION_DEFLATE); // works,lossless compression known as Zip-in-TIFF (1750 msec)
    TIFFSetField(tif, TIFFTAG_COMPRESSION, compressionAlgo); // TODO: check if this is a valid compressoinalgorithm

    // size of strips = nr. of lines taht form a strip
    unsigned int rowsPerStrip  = rowsPerStripArg;
    if (rowsPerStrip <= 0)
      rowsPerStrip = img.GetHeight();
    if (rowsPerStrip>img.GetHeight()) {
      BIASERR("rowsPerStrip="<<rowsPerStrip<<" exceed limit. resetting to img height="<<img.GetHeight());
      rowsPerStrip=img.GetHeight();
    }
    /// \todo TODO: do we need to check that rowsPerStrip is modulo image height?
    BIASASSERT(rowsPerStrip >= 0 );
    BIASASSERT(rowsPerStrip <= img.GetHeight() );
    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);

    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    switch (img.GetChannelCount()){
      case 1:
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK); break;
        //case 2:
      case 3:
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB); break;
        //case 4:
      default: 
        {
          BIASERR("unhandled nr. of channels.");
          return -1;
        }
    }

    //// optional tags
    TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "written Jan Woetzel");
    const string cvsVersion = "CVS $Revision: 2.157 $";
    stringstream swSS;
    swSS <<"BIAS::ImageIO "
      <<"BIAS-"<<BIAS_VERSION_MAJOR<<"."<<BIAS_VERSION_MINOR<<"."<<BIAS_VERSION_PATCH
      <<" "<<cvsVersion<<" (Jan Woetzel)";
    TIFFSetField(tif, TIFFTAG_SOFTWARE, swSS.str().c_str() );

    //TIFFSetField(tif, TIFFTAG_XRESOLUTION, 150.0);
    //TIFFSetField(tif, TIFFTAG_YRESOLUTION, 150.0);
    //TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH); // RESUNIT_CENTIMETER

    // write TIFF in scanlines
    int res=0;
    //if (rowsPerStrip==1)
    //{
    //  // use scanline based API
    //  0000000000000000 these are scanlines
    //  1111111111111111
    //  2222222222222222
    //  for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip ){
    //    void* pd = img.GetImageDataArray()[y];
    //    // 1is returned on success
    //    res = TIFFWriteScanline(tif, pd, y, 0);
    //    BIASASSERT(res==1);
    //  }
    //} 
    //else 
    {
      // use strip based API, possibly encoded
      // 0000000000  these are strips
      // 0000000000
      // 1111111111
      // 1111111111
      // TODO must rowsPerStrip be module height?
      const unsigned int nBytes = rowsPerStrip * img.GetWidth()*img.GetChannelCount()*img.GetDepth() ;
      int strip =0;
      for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip){
        // the number of written Bytes is returned
        res = TIFFWriteEncodedStrip(tif, strip, img.GetImageDataArray()[y], nBytes );
        BIASASSERT((int)res==(int)nBytes);
        strip++;
      }
    }

    TIFFClose(tif);
    return EXIT_SUCCESS;

  } catch (...){
    BIASERR("catched tifflib exception");
    return -1;
  }
}


int BIAS::ImageIO::ImportTIFFLIB(const std::string& FileName, ImageBase& img )
{
  //  /*
  //#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  //  CALLINFO;
  //#endif
  //  try {
  //
  //    TIFF* tif = TIFFOpen(FileName.c_str(), "r");
  //    if (tif==NULL){
  //#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  //      BIASERR("Could not open image for TIFFLIB for reading:" << FileName);
  //#endif
  //      return -1;
  //    }
  //    BIASASSERT(tif!=NULL);
  //
  //    // work:
  //
  //    int width=0, height=0, channelcount=0, bitpersample=0;
  //
  //    // Check that it is of a type that we support
  //
  //
  //
  //      if((== 0) || (bps != 1)){
  //        fprintf(stderr, "Either undefined or unsupported number of bits per sample\n");
  //        exit(42);
  //      }
  //
  //      if((TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &spp) == 0) || (spp != 1)){
  //        fprintf(stderr, "Either undefined or unsupported number of samples per pixel\n");
  //        exit(42);
  //      }
  //
  //    TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &bitpersample) ;
  //
  //
  //    TIFFGetField(tif, TIFFTAG_IMAGEDESCRIPTION, "written by BIAS::ImageIO (Jan Woetzel)");
  //    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      img.GetWidth() );
  //    TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     img.GetHeight() );
  //    TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, img.GetChannelCount() );
  //    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   8*img.GetDepth() ); // more than enough
  //    //TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   img.GetBitDepth() ); // exactly enough
  //    TIFFGetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT); // origin 0,0
  //    // assume planar tiff cfg., needs one channel or interleaved pixel data
  //    BIASASSERT((img.GetChannelCount()==1) || (img.IsInterleaved()) ); 
  //    TIFFGetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG);
  //
  //
  //    // read TIFF in scanlines
  //    int res=0;
  //    //if (rowsPerStrip==1)
  //    //{
  //    //  // use scanline based API
  //    //  0000000000000000 these are scanlines
  //    //  1111111111111111
  //    //  2222222222222222
  //    //  for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip ){
  //    //    void* pd = img.GetImageDataArray()[y];
  //    //    // 1is returned on success
  //    //    res = TIFFReadScanline(tif, pd, y, 0);
  //    //    BIASASSERT(res==1);
  //    //  }
  //    //} 
  //    //else 
  //    {
  //      // use strip based API, possibly encoded
  //      // 0000000000  these are strips
  //      // 0000000000
  //      // 1111111111
  //      // 1111111111
  //      // TODO must rowsPerStrip be module height?
  //      const unsigned int nBytes = rowsPerStrip * img.GetWidth()*img.GetChannelCount()*img.GetDepth() ;
  //      int strip =0;
  //      for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip){
  //        // the number of written Bytes is returned
  //        res = TIFFReadEncodedStrip(tif, strip, img.GetImageDataArray()[y], nBytes );
  //        BIASASSERT((int)res==(int)nBytes);
  //        strip++;
  //      }
  //    }
  //
  //    TIFFClose(tif);
  //    return EXIT_SUCCESS;
  //
  //  } catch (...){
  //    BIASERR("catched tifflib exception");
  //    return -1;
  //  }
  //  */
  return -1;
}
#endif // BIAS_HAVE_TIFF

