#ifndef JAMA_SVD_H
#define JAMA_SVD_H

//#include "tnt/array1d.h"
//#include "tnt/array1d_utils.h"
#include "Matrix.hh"
//#include "tnt/array2d_utils.h"
//#include "tnt/math_utils.h"

namespace BIAS
{
  /** Singular Value Decomposition copied from the JAMA library
      <P>
      For an m-by-n matrix A with m >= n, the singular value decomposition is
      an m-by-n orthogonal matrix U, an n-by-n diagonal matrix S, and
      an n-by-n orthogonal matrix V so that A = U*S*V'.
      <P>
      The singular values, sigma[k] = S[k][k], are ordered so that
      sigma[0] >= sigma[1] >= ... >= sigma[n-1].
      <P>
      The singular value decompostion always exists, so the constructor will
      never fail.  The matrix condition number and the effective numerical
      rank can be computed from this decomposition.

      <p>
      (Adapted from JAMA, a Java Matrix Library, developed by jointly 
      by the Mathworks and NIST; see  http://math.nist.gov/javanumerics/jama).

License for JAMA (see http://math.nist.gov/javanumerics/jama/ copyright notice)

This software is a cooperative product of The MathWorks and the National Institute of Standards and Technology (NIST) which has been released to the public domain. Neither The MathWorks nor NIST assumes any responsibility whatsoever for its use by other parties, and makes no guarantees, expressed or implied, about its quality, reliability, or any other characteristic. 
  */
  template <class T>
  class SVD 
  {


    Matrix<T> U, V,A,VT_;
    Vector<T> s;
    int m, n;
  public:
    T ZeroThreshold_;

    SVD (const Matrix<T> &Arg) {
      using namespace TNT;
      A=Arg;
      ZeroThreshold_=1E-08;
      m = Arg.GetRows();
      n = Arg.GetCols();
      int nu = int((int)min(m,n));
      s = Vector<T>(int((int)min(m+1,n))); 
      U = Matrix<T>(m, nu, T(0));
      V = Matrix<T>(n,n);
      Vector<T> e(n);
      Vector<T> work(m);
  	  Matrix<T> A(Arg);
      int wantu = 1;  					/* boolean */
      int wantv = 1;  					/* boolean */
      int i=0, j=0, k=0;

      // Reduce A to bidiagonal form, storing the diagonal elements
      // in s and the super-diagonal elements in e.

      int nct = int((int)min(m-1,n));
      int nrt = (int)TNT::max(0,(int)min(n-2,m));
      for (k = 0; k < TNT::max(nct,nrt); k++) {
        if (k < nct) {

          // Compute the transformation for the k-th column and
          // place the k-th diagonal in s[k].
          // Compute 2-norm of k-th column without under/overflow.
          s[k] = 0;
          for (i = k; i < m; i++) {
            s[k] = _hypot(s[k],A[i][k]);
          }
          if (s[k] != 0.0) {
            if (A[k][k] < 0.0) {
              s[k] = -s[k];
            }
            for (i = k; i < m; i++) {
              A[i][k] /= s[k];
            }
            A[k][k] += 1.0;
          }
          s[k] = -s[k];
        }
        for (j = k+1; j < n; j++) {
          if ((k < nct) && (s[k] != 0.0))  {

            // Apply the transformation.

            double t = 0;
            for (i = k; i < m; i++) {
              t += A[i][k]*A[i][j];
            }
            t = -t/A[k][k];
            for (i = k; i < m; i++) {
              A[i][j] += (T)t*A[i][k];
            }
          }

          // Place the k-th row of A into e for the
          // subsequent calculation of the row transformation.

          e[j] = A[k][j];
        }
        if (wantu & (k < nct)) {

          // Place the transformation in U for subsequent back
          // multiplication.

          for (i = k; i < m; i++) {
            U[i][k] = A[i][k];
          }
        }
        if (k < nrt) {

          // Compute the k-th row transformation and place the
          // k-th super-diagonal in e[k].
          // Compute 2-norm without under/overflow.
          e[k] = 0;
          for (i = k+1; i < n; i++) {
            e[k] = _hypot(e[k],e[i]);
          }
          if (e[k] != 0.0) {
            if (e[k+1] < 0.0) {
              e[k] = -e[k];
            }
            for (i = k+1; i < n; i++) {
              e[i] /= e[k];
            }
            e[k+1] += 1.0;
          }
          e[k] = -e[k];
          if ((k+1 < m) & (e[k] != 0.0)) {

            // Apply the transformation.

            for (i = k+1; i < m; i++) {
              work[i] = 0.0;
            }
            for (j = k+1; j < n; j++) {
              for (i = k+1; i < m; i++) {
                work[i] += e[j]*A[i][j];
              }
            }
            for (j = k+1; j < n; j++) {
              double t = -e[j]/e[k+1];
              for (i = k+1; i < m; i++) {
                A[i][j] += (T)t*work[i];
              }
            }
          }
          if (wantv) {

            // Place the transformation in V for subsequent
            // back multiplication.

            for (i = k+1; i < n; i++) {
              V[i][k] = e[i];
            }
          }
        }
      }

      // Set up the final bidiagonal matrix or order p.

      int p = (int)min(n,m+1);
      if (nct < n) {
        s[nct] = A[nct][nct];
      }
      if (m < p) {
        s[p-1] = 0.0;
      }
      if (nrt+1 < p) {
        e[nrt] = A[nrt][p-1];
      }
      e[p-1] = 0.0;

      // If required, generate U.

      if (wantu) {
        for (j = nct; j < nu; j++) {
          for (i = 0; i < m; i++) {
            U[i][j] = 0.0;
          }
          U[j][j] = 1.0;
        }
        for (k = nct-1; k >= 0; k--) {
          if (s[k] != 0.0) {
            for (j = k+1; j < nu; j++) {
              double t = 0;
              for (i = k; i < m; i++) {
                t += U[i][k]*U[i][j];
              }
              t = -t/U[k][k];
              for (i = k; i < m; i++) {
                U[i][j] += (T)t*U[i][k];
              }
            }
            for (i = k; i < m; i++ ) {
              U[i][k] = -U[i][k];
            }
            U[k][k] = (T)1.0 + U[k][k];
            for (i = 0; i < k-1; i++) {
              U[i][k] = 0.0;
            }
          } else {
            for (i = 0; i < m; i++) {
              U[i][k] = 0.0;
            }
            U[k][k] = 1.0;
          }
        }
      }

      // If required, generate V.

      if (wantv) {
        for (k = n-1; k >= 0; k--) {
          if ((k < nrt) & (e[k] != 0.0)) {
            for (j = k+1; j < nu; j++) {
              double t = 0;
              for (i = k+1; i < n; i++) {
                t += V[i][k]*V[i][j];
              }
              t = -t/V[k+1][k];
              for (i = k+1; i < n; i++) {
                V[i][j] += (T)t*V[i][k];
              }
            }
          }
          for (i = 0; i < n; i++) {
            V[i][k] = 0.0;
          }
          V[k][k] = 1.0;
        }
      }

      // Main iteration loop for the singular values.

      int pp = p-1;
      int iter = 0;
      double eps = pow(2.0,-52.0);
      while (p > 0) {
        int k=0;
        int kase=0;

        // Here is where a test for too many iterations would go.

        // This section of the program inspects for
        // negligible elements in the s and e arrays.  On
        // completion the variables kase and k are set as follows.

        // kase = 1     if s(p) and e[k-1] are negligible and k<p
        // kase = 2     if s(k) is negligible and k<p
        // kase = 3     if e[k-1] is negligible, k<p, and
        //              s(k), ..., s(p) are not negligible (qr step).
        // kase = 4     if e(p-1) is negligible (convergence).

        for (k = p-2; k >= -1; k--) {
          if (k == -1) {
            break;
          }
          if (TNT::abs(e[k]) <= eps*(TNT::abs(s[k]) + TNT::abs(s[k+1]))) {
            e[k] = 0.0;
            break;
          }
        }
        if (k == p-2) {
          kase = 4;
        } else {
          int ks;
          for (ks = p-1; ks >= k; ks--) {
            if (ks == k) {
              break;
            }
            double t = (ks != p ? TNT::abs(e[ks]) : 0.) + 
              (ks != k+1 ? TNT::abs(e[ks-1]) : 0.);
            if (TNT::abs(s[ks]) <= eps*t)  {
              s[ks] = 0.0;
              break;
            }
          }
          if (ks == k) {
            kase = 3;
          } else if (ks == p-1) {
            kase = 1;
          } else {
            kase = 2;
            k = ks;
          }
        }
        k++;

        // Perform the task indicated by kase.

        switch (kase) {

          // Deflate negligible s(p).

        case 1: {
          double f = e[p-2];
          e[p-2] = 0.0;
          for (j = p-2; j >= k; j--) {
            double t = _hypot(s[j],f);
            double cs = s[j]/t;
            double sn = f/t;
            s[j] = (T)t;
            if (j != k) {
              f = -sn*e[j-1];
              e[j-1] = (T)cs*e[j-1];
            }
            if (wantv) {
              for (i = 0; i < n; i++) {
                t = cs*V[i][j] + sn*V[i][p-1];
                V[i][p-1] = (T)-sn*V[i][j] + (T)cs*V[i][p-1];
                V[i][j] = (T)t;
              }
            }
          }
        }
          break;

          // Split at negligible s(k).

        case 2: {
          double f = e[k-1];
          e[k-1] = 0.0;
          for (j = k; j < p; j++) {
            double t = _hypot(s[j],f);
            double cs = s[j]/t;
            double sn = f/t;
            s[j] = (T)t;
            f = -sn*e[j];
            e[j] = (T)cs*e[j];
            if (wantu) {
              for (i = 0; i < m; i++) {
                t = cs*U[i][j] + sn*U[i][k-1];
                U[i][k-1] = (T)-sn*U[i][j] + (T)cs*U[i][k-1];
                U[i][j] = (T)t;
              }
            }
          }
        }
          break;

          // Perform one qr step.

        case 3: {

          // Calculate the shift.
   
          double scale = TNT::max(TNT::max(TNT::max(TNT::max(
                                                             TNT::abs(s[p-1]),TNT::abs(s[p-2])),TNT::abs(e[p-2])), 
                                           TNT::abs(s[k])),TNT::abs(e[k]));
          double sp = s[p-1]/scale;
          double spm1 = s[p-2]/scale;
          double epm1 = e[p-2]/scale;
          double sk = s[k]/scale;
          double ek = e[k]/scale;
          double b = ((spm1 + sp)*(spm1 - sp) + epm1*epm1)/2.0;
          double c = (sp*epm1)*(sp*epm1);
          double shift = 0.0;
          if ((b != 0.0) || (c != 0.0)) {
            shift = sqrt(b*b + c);
            if (b < 0.0) {
              shift = -shift;
            }
            shift = c/(b + shift);
          }
          double f = (sk + sp)*(sk - sp) + shift;
          double g = sk*ek;
   
          // Chase zeros.
   
          for (j = k; j < p-1; j++) {
            double t = _hypot(f,g);
            double cs = f/t;
            double sn = g/t;
            if (j != k) {
              e[j-1] = (T)t;
            }
            f = cs*s[j] + sn*e[j];
            e[j] = cs*e[j] - sn*s[j];
            g = sn*s[j+1];
            s[j+1] = (T)cs*s[j+1];
            if (wantv) {
              for (i = 0; i < n; i++) {
                t = (T)cs*V[i][j] + (T)sn*V[i][j+1];
                V[i][j+1] = (T)-sn*V[i][j] + (T)cs*V[i][j+1];
                V[i][j] = (T)t;
              }
            }
            t = _hypot(f,g);
            cs = f/t;
            sn = g/t;
            s[j] = (T)t;
            f = cs*e[j] + sn*s[j+1];
            s[j+1] = (T)-sn*e[j] + (T)cs*s[j+1];
            g = sn*e[j+1];
            e[j+1] = (T)cs*e[j+1];
            if (wantu && (j < m-1)) {
              for (i = 0; i < m; i++) {
                t = cs*U[i][j] + sn*U[i][j+1];
                U[i][j+1] = (T)-sn*U[i][j] + (T)cs*U[i][j+1];
                U[i][j] = (T)t;
              }
            }
          }
          e[p-2] = (T)f;
          iter = iter + 1;
        }
          break;

          // Convergence.

        case 4: {

          // Make the singular values positive.
   
          if (s[k] <= 0.0) {
            s[k] = (T)(s[k] < 0.0 ? -s[k] : 0.0);
            if (wantv) {
              for (i = 0; i <= pp; i++) {
                V[i][k] = -V[i][k];
              }
            }
          }
   
          // Order the singular values.
   
          while (k < pp) {
            if (s[k] >= s[k+1]) {
              break;
            }
            double t = s[k];
            s[k] = s[k+1];
            s[k+1] = (T)t;
            if (wantv && (k < n-1)) {
              for (i = 0; i < n; i++) {
                t = V[i][k+1]; V[i][k+1] = V[i][k]; V[i][k] = (T)t;
              }
            }
            if (wantu && (k < m-1)) {
              for (i = 0; i < m; i++) {
                t = U[i][k+1]; U[i][k+1] = U[i][k]; U[i][k] = (T)t;
              }
            }
            k++;
          }
          iter = 0;
          p--;
        }
          break;
        }
      }
	  VT_=V.Transpose();
    }


    void GetU (Matrix<T> &A) 
    {
      using namespace TNT;
   	  int minm = (int)min(m+1,n);

      A = Matrix<T>(m, minm);

      for (int i=0; i<m; i++)
        for (int j=0; j<minm; j++)
          A[i][j] = U[i][j];
   	
    }

    Matrix<T> GetU()
    { Matrix<T> U; GetU(U); return U; }

    /* Return the right singular vectors */

    void GetV (Matrix<T> &A) 
    {
   	  A = V;
    }

    Matrix<T> GetV() { return V; }

	void GetVT (Matrix<T> &A) 
    {
   	  A = VT_;
    }

    Matrix<T> GetVT() { return VT_; }

    /** Return the one-dimensional array of singular values */

    void getSingularValues (Vector<T> &x) 
    {
      x = s;
    }

    /** Return the diagonal matrix of singular values
        @return     S
    */

    void GetS (Matrix<T> &A) {
   	  A = Matrix<T>(n,n);
      for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++) {
          A[i][j] = 0.0;
        }
        A[i][i] = s[i];
      }
    }

    Vector<T> GetS () { return s;  }

    /** Two norm  (TNT::max(S)) */

    double norm2 () {
      return s[0];
    }

    /** Two norm of condition number (TNT::max(S)/(int)min(S)) */

    double cond () {
      using namespace TNT;
      return s[0]/s[(int)min(m,n)-1];
    }

    /** Effective numerical matrix rank
        @return     Number of nonnegligible singular values.
    */

    int rank () 
    {
      double eps = pow(2.0,-52.0);
      double tol = TNT::max(m,n)*s[0]*eps;
      int r = 0;
      for (int i = 0; i < s.dim(); i++) {
        if (s[i] > tol) {
          r++;
        }
      }
      return r;
    }


    /**	@brief returns dim of nullspace using the absolute threshold criterion
        
    For rectangular matrices this is the number of singular values which 
    are (about) zero. The nullspace is spaned by the singular vectors 
    corresponding to the zero singular values.
    woelk 4 2003: changed API, diffentiate between Left and Right Nullspace
    @author Jan Woetzel 04/04/2002    **/
    const int AbsNullspaceDim() const
    {
      int dim = 0;
      int index = s.size()-1; 

      //cerr << "ZeroThreshold_: "<<ZeroThreshold_<<endl;
      while (index >= 0 && fabs(s[index]) < ZeroThreshold_ ) {
        // singular values are sorted in descending order is still (around zero
        index--;
        dim++;
      }
  
      if (A.num_cols()>A.num_rows()){
        dim+=A.num_cols()-A.num_rows();
      }
      return dim;
    }


    bool GetNullvector(Vector<T>& NullVec,
                       const int last_offset  ) 
    {
      // create a new vector
      Matrix<T> VT=V.Transpose();
      NullVec.newsize(VT.num_cols());
      // fill the (null-)vector

      for (int col=0; col < VT.num_cols(); col++) {
        NullVec[col] = VT[ VT.num_cols() -1 - last_offset ][col];
      }
    
      // check if there is a nullvector with this last_offset_index
      if (last_offset > (AbsNullspaceDim() -1)) {
        // index not OK
        return false;
      } else {
        // index OK      
        return true;
      }
    }

    Vector<T> GetNullvector(const int last_offset=0) 
    {
      Vector<T> vec(V.Transpose().num_cols(), 0.0);
      if (!GetNullvector(vec, last_offset)) {
        BIASERR("No Nullvector for "<<last_offset);
      }
      return vec;
    }


    Matrix<T> Invert()
    {
      Matrix<T> VT=V.Transpose();
      Matrix<T> V(VT.num_cols(), VT.num_rows());
      Matrix<T> UT(U.num_cols(), U.num_rows());
      Matrix<T> SdT(VT.num_rows(), U.num_cols());
      SdT.SetZero();
      for (register int i = 0; i < s.size(); i++){
        SdT[i][i] = (s[i] < ZeroThreshold_ && s[i] > -ZeroThreshold_)
          ? (0.0) : (1.0/s[i]);
      }
      V = VT.Transpose();
      UT = U.Transpose();

      return V * SdT * UT;
    }

  };
  
}


inline int SVD_Solve(BIAS::Matrix<double>& A, BIAS::Vector<double>& b, BIAS::Vector<double>& x)
{
	using namespace BIAS;
  int res = 0;
  // make squared symmetric matrix A^T * A
  Matrix<double> ATA(A.num_cols(), A.num_cols());
  //ATA = A.Transpose()*A;
  A.GetSystemMatrix(ATA);

  // compute ATA = U * S * V^T
  SVD<double> svd(ATA);
  Vector<double> d(A.num_rows()), z(A.num_cols());
  Vector<double> S = svd.GetS();
//  cerr<<"S: "<<S<<endl; 
  
  // A * x = b  => x = (A^T * A )^-1 * A^T * b
  //                 = (U * S * V^T)^-1 * A^T * b 
  // by symmetry we know: A^T * A = (A^T * A)^T 
  //                 = (V * S * U^T)^-1 * A^T * b 
  // since U and V^T are square orthogonal matrices here
  //                 = U^-T * S^-1 * V^T * A^T * b
  //                 = U    * S^-1 *       d
  d = svd.GetVT() * A.Transpose() * b;
  
  // compute z = S^-1 * d
  //BIASGDOUT(D_SVD_SOLVE, "d: "<<d<<endl);
  
    for (int i=0; i<A.num_cols(); i++) {
      if (fabs(S[i])>svd.ZeroThreshold_) {
        z[i] = d[i]/S[i];
      } else {
        z[i] = 0.0;
        res--;
      }
    }
  

  if (res<0) {
	  std::cerr<<"Error in SVD, degenerate case in Solve ("
            << -res <<" dim. solution space), S=" 
			<< S<<std::endl;
  }
  
  // get final solution x = U * z
  x = svd.GetU() * z;
  return res;
}

#endif
// JAMA_SVD_H
